"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.jsonThink = exports.JSONThink = exports.isTransformDescription = exports.isObjectKeyType = exports.withFromJSONThink = exports.withToJSONThinkWithValue = exports.withToJSONThink = void 0;
var BigIntDescriptor_1 = require("./descriptors/BigIntDescriptor");
var BufferDescriptor_1 = require("./descriptors/BufferDescriptor");
var DateDescriptor_1 = require("./descriptors/DateDescriptor");
var pipe = function (vale) {
    var currentValue = vale;
    var self = {
        pipe: function (fn) {
            currentValue = fn(currentValue);
            return this;
        },
        valueOf: function () {
            return currentValue;
        },
    };
    return self;
};
var withToJSONThink = function (value) {
    return typeof value === "object" ? typeof value.toJSONThink === "function" : false;
};
exports.withToJSONThink = withToJSONThink;
var withToJSONThinkWithValue = function (value) {
    return typeof value === "object" ? typeof value.toJSONThink === "function" : false;
};
exports.withToJSONThinkWithValue = withToJSONThinkWithValue;
var withFromJSONThink = function (value) {
    var typeValue = typeof value;
    return typeValue === "object" || typeValue === "function"
        ? typeof value.fromJSONThink === "function"
        : false;
};
exports.withFromJSONThink = withFromJSONThink;
var isObjectKeyType = function (value, propTypeName) {
    if (propTypeName === void 0) { propTypeName = "$$type"; }
    return typeof value === "object" ? typeof value[propTypeName] === "string" : false;
};
exports.isObjectKeyType = isObjectKeyType;
var isTransformDescription = function (value) {
    return typeof value === "object"
        ? typeof value.test === "function" &&
            exports.withToJSONThinkWithValue(value) &&
            exports.withFromJSONThink(value)
        : false;
};
exports.isTransformDescription = isTransformDescription;
var globalDescriptores = new Map();
globalDescriptores.set("Date", DateDescriptor_1.DateDescriptor);
globalDescriptores.set("BigInt", BigIntDescriptor_1.BigIntDescriptor);
globalDescriptores.set("Buffer", BufferDescriptor_1.BufferDescriptor);
var JSONThink = /** @class */ (function () {
    function JSONThink(options) {
        var _this = this;
        var _a, _b;
        this.options = options;
        this.propTypeName = (_b = (_a = this.options) === null || _a === void 0 ? void 0 : _a.propTypeName) !== null && _b !== void 0 ? _b : "$$type";
        this.listThinks = new Map();
        this.listThinksByName = new Map();
        this.listThinksDescriptions = globalDescriptores;
        this.replacerLow = function (replacer) {
            return function (key, value) {
                var _a, _b;
                if (exports.withToJSONThink(value)) {
                    var res = value.toJSONThink();
                    return __assign(__assign({}, res), (_a = {}, _a[_this.propTypeName] = _this.listThinks.get(value.constructor), _a));
                }
                if (_this.listThinksDescriptions.size) {
                    var iterator = _this.listThinksDescriptions.entries();
                    while (true) {
                        var _c = iterator.next(), valueIterator = _c.value, done = _c.done;
                        if (done)
                            break;
                        var _d = __read(valueIterator, 2), descName = _d[0], desc = _d[1];
                        if (desc.test(value)) {
                            var res = desc.toJSONThink(value);
                            return __assign(__assign({}, res), (_b = {}, _b[_this.propTypeName] = descName, _b));
                        }
                    }
                }
                return typeof replacer === "function" ? replacer(key, value) : value;
            };
        };
    }
    JSONThink.prototype.use = function (ref, name) {
        var nameValue = name !== null && name !== void 0 ? name : ref.name;
        this.listThinksByName.set(nameValue, ref);
        if (exports.isTransformDescription(ref)) {
            this.listThinksDescriptions.set(nameValue, ref);
        }
        else {
            this.listThinks.set(ref, nameValue);
        }
        return this;
    };
    JSONThink.prototype.stringify = function (value, replacer, space) {
        var nextObj = this.deep(value, "", this.replacerLow(replacer));
        return JSON.stringify(nextObj, null, space);
    };
    JSONThink.prototype.deep = function (value, key, replacer) {
        var _this = this;
        var res = pipe(value)
            .pipe(function (v) { var _a; return (_a = replacer === null || replacer === void 0 ? void 0 : replacer(key, value)) !== null && _a !== void 0 ? _a : v; })
            .pipe(function (v) { var _a, _b; return (typeof v === "object" ? (_b = (_a = v.toJSON) === null || _a === void 0 ? void 0 : _a.call(v)) !== null && _b !== void 0 ? _b : v : v); })
            .valueOf();
        if (Array.isArray(res)) {
            return res.map(function (value, key) {
                return _this.deep(value, key.toString(), replacer);
            });
        }
        if (typeof res === "object") {
            var re = Object.entries(res).reduce(function (a, _a) {
                var _b;
                var _c = __read(_a, 2), key = _c[0], value = _c[1];
                return __assign(__assign({}, a), (_b = {}, _b[key] = _this.deep(value, key, replacer), _b));
            }, {});
            return re;
        }
        return res;
    };
    JSONThink.prototype.parse = function (text, reviver) {
        var _this = this;
        var reviverP = function (key, value) {
            var _a;
            if (exports.isObjectKeyType(value, _this.propTypeName)) {
                var typeKey = value[_this.propTypeName];
                var typeCls = _this.listThinksByName.get(typeKey);
                if (exports.withFromJSONThink(typeCls)) {
                    return typeCls.fromJSONThink(value);
                }
            }
            return (_a = reviver === null || reviver === void 0 ? void 0 : reviver(key, value)) !== null && _a !== void 0 ? _a : value;
        };
        return JSON.parse(text, reviverP);
    };
    return JSONThink;
}());
exports.JSONThink = JSONThink;
exports.jsonThink = new JSONThink();
